#include <limits.h>
#include "abr.h"

ABR creaABR ()
{
  return EMPTY_TREE;
}


void distruggeABR (ABR *pT)
{
  if (*pT != EMPTY_TREE)
  {
    distruggeABR(&(*pT)->Ts);
    distruggeABR(&(*pT)->Td);
    free(*pT);
    *pT = EMPTY_TREE;
  }
}


boolean ABRvuoto (ABR T)
{
  return (T == EMPTY_TREE);
}


boolean memberABR (int i, ABR T)
{
#ifdef RICORSIVA
  /* Implementazione ricorsiva */
  if (T == EMPTY_TREE)
    return FALSE;
  else if (i == T->val)
    return TRUE;
  else if (i < T->val)
    return memberABR(i,T->Ts);
  else /* if (i > T->val) */
    return memberABR(i,T->Td);
#else
  /* Implementazione iterativa */
  while ( (T != EMPTY_TREE) && (i != T->val) )
    if (i < T->val)
      T = T->Ts;
    else /* if (i > T->val) */
      T = T->Td;

  return (T != EMPTY_TREE);
#endif
}


ABR creanodoABR (int i)
{
  nodo *r;

  /* Alloca l'elemento radice */
  r = (nodo *) malloc(sizeof(nodo));
  if (r == NULL)
  {
    fprintf(stderr,"Errore nell'allocazione di un nodo!\n");
    exit(EXIT_FAILURE);
  }

  r->val = i;
  r->padre = NULL;
  r->Ts = NULL;
  r->Td = NULL;

  return r;
}


ABR insertABR (int i, ABR T)
{
  if (T == EMPTY_TREE)
    T = creanodoABR(i);
  else if (i < T->val)
  {
    T->Ts = insertABR(i,T->Ts);
    T->Ts->padre = T;
  }
  else if (i > T->val)
  {
    T->Td = insertABR(i,T->Td);
    T->Td->padre = T;
  }

  return T;
}


/* Appende direttamente l'unico sottoalbero del nodo n dell'ABR T
   al nodo padre e dealloca il nodo n */
ABR toglienodoABR (nodo *n, ABR T)
{
  ABR Tf;


  /* Determina l'unico sottoalbero non vuoto del nodo n (eventualmente EMPTY_TREE) */
  Tf = ( (n->Ts == EMPTY_TREE) ? n->Td : n->Ts );

  /* Se esiste un sottoalbero, lo appende al padre (eventualmente NULL) */
  if (Tf != EMPTY_TREE) Tf->padre = n->padre;

  /* Se n e' la radice, sostituisce l'albero T con il sottoalbero */
  if (n == T)
    T = Tf;
  else /* Se n ha un padre, determina in quale sottoalbero del padre stava
          e sostituisce il sottoalbero a se stesso */
  {
    if (n == n->padre->Ts)
      n->padre->Ts = Tf;
    else if (n == n->padre->Td)
      n->padre->Td = Tf;
  }

  /* Dealloca il nodo */
  free(n);

  return T;
}


/* Determina il nodo che corrisponde al valore massimo dell'ABR T */
nodo *argmaxABR (ABR T)
{
  if (T == EMPTY_TREE) return NULL;
  while (T->Td != EMPTY_TREE)
    T = T->Td;
  return T;
}


ABR deleteABR (int i, ABR T)
{
  nodo *n;

  /* Caso base 1: l'albero e' vuoto (o i non appartiene a T), non fa nulla */
  if (T != EMPTY_TREE)
  {
    /* I due casi ricorsivi */
    if (i < T->val)
      T->Ts = deleteABR(i,T->Ts);
    else if (i > T->val)
      T->Td = deleteABR(i,T->Td);
    /* Caso base 2: i e' la radice e ha al massimo un figlio */
    else if ( (T->Ts == EMPTY_TREE) || (T->Td == EMPTY_TREE) )
      /* Appende il sottoalbero al padre e cancella la radice */
      T = toglienodoABR(T,T);
    /* Caso base 3: i e' la radice e ha due figli */
    else
    {
      /* Trova l'elemento massimo del sottoalbero sinistro
         sovrascrive la radice con esso e lo cancella */
      n = argmaxABR(T->Ts);
      T->val = n->val;
      T = toglienodoABR(n,T);
    }
  }

  return T;
}


int minABR (ABR T)
{
#ifdef RICORSIVA
  /* Implementazione ricorsiva */
  if (T == EMPTY_TREE)
    return INT_MAX;
  else if (T->Ts == EMPTY_TREE)
    return T->val;
  else
    return minABR(T->Ts);
#else
  /* Implementazione iterativa */
  if (T == EMPTY_TREE) return INT_MAX;
  while (T->Ts != EMPTY_TREE)
    T = T->Ts;

  return T->val;
#endif
}


int maxABR (ABR T)
{
#ifdef RICORSIVA
  /* Implementazione ricorsiva */
  if (T == EMPTY_TREE)
    return INT_MIN;
  else if (T->Td == EMPTY_TREE)
    return T->val;
  else
    return maxABR(T->Td);
#else
  /* Implementazione iterativa */
  if (T == EMPTY_TREE) return INT_MIN;
  while (T->Td != EMPTY_TREE)
    T = T->Td;

  return T->val;
#endif
}


void stampaABR (ABR T)
{
  if (T != EMPTY_TREE)
  {
    stampaABR(T->Ts);
    printf(" %d",T->val);
    stampaABR(T->Td);
  }
}
